from datetime import datetime
import sys
import os

from constants import ORDER_SIDE, Side
import chardet



#没有融券
def encode_ims_side(side):
    rtn_trade_side = 0
    if side == "Long Buy":
        rtn_trade_side = 3
    elif side == "Long Sell":
        rtn_trade_side = 6
    elif side == "Short Sell":
        rtn_trade_side = 4
    elif side == "Buy To Cover":
        rtn_trade_side = 5
    return rtn_trade_side

def decode_ims_side(side):
    order_side = ""
    if side == 3:
        order_side = "Long Buy"
    elif side == 6:
        order_side = "Long Sell"
    elif side == 4:
        order_side = "Short Sell"
    elif side == 5:
        order_side = "Buy To Cover"
    return order_side

def decode_atx_target_type(side):
    f_side = 0
    if side in [1,5]:
        return 1
    elif side == 2:
        return 2

def decode_ims_contract_type(contract_type):
    str_type = ""
    if contract_type == 0:
        str_type = "TRS"
    elif contract_type == 1:
        str_type = "RISKY"
    return str_type

def decode_ims_status(status):
    new_status = ""
    if status == '未成交' or status == '部分成交':
        new_status = 'active'
    elif status == '全部成交':
        new_status = 'filled'
    elif status == 'CANCELED':
        new_status = 'canceled'
    return new_status

def side_to_target_type(side):
    target_type = ""
    if side in [1,5]:
        target_type = "buy"
    elif side == 2:
        target_type = "sell"
    return target_type

def decode_exchange_id(exchange):
    if exchange.startswith('6'):
        exchange = 101
    elif exchange.startswith('0'):
        exchange = 102
    else:
        exchange = 102

    return exchange

def decode_ims_market(stock_code):
    exchange = ""
    if stock_code.startswith('6'):
        exchange = 0
    elif stock_code.startswith('0'):
        exchange = 1
    else:
        exchange = 1

    return exchange 


def get_log_default_path():
    # python2: linux2, python3: linux
    if sys.platform.startswith("linux") or sys.platform == "darwin":
        dirs = "/shared/log"
    elif sys.platform == "win32":
        dirs = os.path.join(get_windows_first_disk() + "/tmp/linker/log")
    else:
        dirs = '.'

    return dirs

def get_today_date():
    return datetime.today().strftime('%Y-%m-%d')

def get_digit_from_env(env_name, default_num):
    num = str(os.environ.get(env_name))
    return int(num) if num.isdigit() else default_num

def get_log_given_path(path):
    dirs = os.path.join(path)
    return path

def decode_ordtype(order_type:str):
    ord_type = -1
    mapping = {
        'TWAP': 0,
        'VWAP': 1,
        'VOLINLINE': 2,
        'ICEBERG': 3
    }
    ord_type = mapping.get(order_type, -1)
    return ord_type

def check_charset(file_path):
    with open(file_path, "rb") as f:
        data = f.read(4)
        charset = chardet.detect(data)['encoding']
    return charset




from datetime import datetime, timedelta
import chardet
import argparse


def encode_gf_side(side):
    rtn_trade_side = 0
    if side == "0B":
        rtn_trade_side = 1
    elif side == "0S":
        rtn_trade_side = 2
    elif side == "4B":
        rtn_trade_side = 3
    return rtn_trade_side


def decode_gf_status(status):
    new_status = ""
    if status in ['2', '7']:
        new_status = 'active'
    elif status == '8':
        new_status = 'filled'
    elif status in ['5', '6']:
        new_status = 'canceled'
    return new_status

def side_to_target_type(side):
    target_type = ""
    if side == 1:
        target_type = "buy"
    elif side == 2:
        target_type = "sell"
    return target_type

def decode_exchange_id(exchange):
    if exchange.startswith('6'):
        exchange = 101
    elif exchange.startswith('0'):
        exchange = 102
    else:
        exchange = 102

    return exchange

def encode_gf_market(stock_code):
    exchange = ""
    if stock_code.startswith('6'):
        exchange = '1'
    elif stock_code.startswith('5'):
        exchange = '1'
    elif stock_code.startswith('0'):
        exchange = '0'
    else:
        exchange = '0'

    return exchange

"""
上海'1'，深圳'0'，沪港通 '8'，深圳通'G'，中金'F'，上期'H'，郑商'Z'，大商'D
"""
def decode_gf_market(market):
    exchange = 0
    if market == '0':
        exchange = 102
    elif market == '1':
        exchange = 101
    return exchange

def get_today_date():
    return datetime.today().strftime('%Y%m%d')

def decode_ordtype(order_type:str):
    ord_type = -1
    mapping = {
        'vwap': '1001',
        'twap': '1002',
        'kf_vwap_plus': '3001',
        'kf_twap_plus': '3002',
        'ft-wap-ai': '4001',
        'ft-wap-ai-plus' : '4003'
    }
    ord_type = mapping.get(order_type, -1)
    return ord_type

def check_charset(file_path):
    with open(file_path, "rb") as f:
        data = f.read(4)
        charset = chardet.detect(data)['encoding']
    return charset

def get_time_from_str(time_str):
    
    today = datetime.now().date()
    datetime_str = f"{today} {time_str}"
    dt = datetime.strptime(datetime_str, "%Y-%m-%d %H:%M:%S")
    utc_dt = dt - timedelta(hours=8)
    
    return utc_dt

def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


from datetime import datetime, timezone, timedelta
import os
import queue
import threading
import time
import json
import signal
import sys
import traceback
import argparse
import csv
import dbf
import shutil
import pathlib


from concurrent.futures import ThreadPoolExecutor
from pymongo import MongoClient, ASCENDING, DESCENDING

from ..mongo_client_gateway import MongoClientTradeGateway
from ..utils import (get_exchange_from_ticker, check_today_index)
from .constants import (FILES, GATEWAY_NAME)
from .utils import (side_to_target_type, decode_exchange_id, decode_ordtype, decode_cats_status, decode_cats_side)
from ..logger import Logger



try:
    import thread
except ImportError:
    import _thread as thread

LL9 = 1000000000


class CatServer(MongoClientTradeGateway):
    def __init__(self, config_filename, endtime):
        MongoClientTradeGateway.__init__(self, config_filename, endtime, GATEWAY_NAME)

        self.insert_orders = {}
        self.oid_to_ref_list = {}
        self.sids = []
        self.oid_to_traded = {}
        self.oid_to_traded_money = {}
        self.sid_to_req = {}
        self.oid_to_req = {}
        self.oid_to_mid = {}
        self.oid_to_acc = {}
        self.db_id_to_oid = {}
        self.ref_to_oid = {}
        self.order_index = -1 #dbf文件当前index
        self.trade_index = -1
        self.pos_index = -1
        self.oid_to_local_ids = {}
        self.acct_type = {}
        self.client_id = {}
        self.oid_to_algoid = {}
        self.algoid_to_oid = {}
        self.trade_index = -1
        self.order_index = -1
        self.algo_index = -1
        self.load_tg_setting(config_filename)
        self.init_dbf_tables()

    def monitor(self):
        self.thread_pool.submit(self.monitor_algo_order_insert)
        #self.thread_pool.submit(self.monitor_cancel_order_insert)
        self.thread_pool.submit(self.monitor_algo_order_update)
        self.thread_pool.submit(self.monitor_order_update)
        self.thread_pool.submit(self.monitor_trade_update)
        #self.thread_pool.submit(self.monitor_dbf_asset)
        self.thread_pool.submit(self.monitor_dbf_pos)
        self.thread_pool.submit(self.date_change)

    def load_tg_setting(self, config_filename):
        try:
            f = open(config_filename, encoding='gbk')
            setting = json.load(f)
            #self.insert_order_path = setting['insert_order_path'].replace('/', '\\')
            self.recv_msg_dir = setting['recv_msg_path'].replace('/', '\\')
            self.recv_msg_trade_dir = setting['recv_msg_path_trade'].replace('/', '\\')
            self.insert_order_path = setting['insert_order_path'].replace('/', '\\')
            for acc in self.accounts_run:
                config = self.config[acc]
                self.acct_type[acc] = config['acct_type']
                self.client_id[acc] = config['client_id']
        except Exception as e:
            err = traceback.format_exc()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback, limit=None, file=sys.stdout)
            print(f"load config failed! (exception){err}")
            exit(0)
    @MongoClientTradeGateway.error_handler
    def init_dbf_tables(self):
        #not use
        pos_filename = self.recv_msg_dir + '\\' + FILES.POSITIONS  + '.dbf'
        self.position_order_table = dbf.Table(filename=pos_filename, codepage='utf8', on_disk=True)
        self.position_order_table.open(mode=dbf.READ_WRITE)
        self.logger.info(f"[init_dbf_tables open] (filename){pos_filename}")
        order_filename = self.recv_msg_trade_dir + '\\' + FILES.ORDERS + '.dbf'
        rtn_order_table = dbf.Table(filename=order_filename,codepage='utf8', on_disk=True)
        with rtn_order_table.open(mode=dbf.READ_WRITE):
            if len(rtn_order_table) <= 0:
                self.logger.warning(f'no record in {order_filename}')
            else:
                while rtn_order_table[self.order_index+1] is not rtn_order_table.last_record:
                    self.order_index += 1
                    record = rtn_order_table[self.order_index]
                    write_time = str(record.WRITE_TIME).strip(' ')[:-4]
                    if check_today_index(write_time):
                        break

        algo_order_filename = self.recv_msg_dir + '\\' + FILES.ALGO_ORDERS + '.dbf'
        rtn_algo_order_table = dbf.Table(filename=algo_order_filename,codepage='utf8', on_disk=True)
        with rtn_algo_order_table.open(mode=dbf.READ_WRITE):
            if len(rtn_algo_order_table) <= 0:
                self.logger.warning(f'no record in {algo_order_filename}')
            else:
                while rtn_algo_order_table[self.algo_index+1] is not rtn_algo_order_table.last_record:
                    self.algo_index += 1
                    record = rtn_algo_order_table[self.algo_index]
                    write_time = str(record.WRITE_TIME).split('@')[1].strip(' ')[:-4]
                    if check_today_index(write_time):
                        break

        trade_filename = self.recv_msg_trade_dir + '\\' + FILES.TRADES + '.dbf'
        rtn_trade_table = dbf.Table(filename=trade_filename,codepage='utf8', on_disk=True)
        with rtn_trade_table.open(mode=dbf.READ_WRITE):
            if len(rtn_trade_table) <= 0:
                self.logger.warning(f'no record in {trade_filename}')
            else:
                while rtn_trade_table[self.trade_index+1] is not rtn_trade_table.last_record:
                    self.trade_index += 1
                    record = rtn_trade_table[self.trade_index]
                    write_time = str(record.TIME).strip(' ')[:-4]
                    if check_today_index(write_time):
                        break

    @MongoClientTradeGateway.error_handler
    def str2bool(self, v: str):
        if isinstance(v, bool):
            return v
        if v.lower() in ('yes', 'true', 't', 'y', '1'):
            return True
        elif v.lower() in ('no', 'false', 'f', 'n', '0'):
            return False
        else:
            raise argparse.ArgumentTypeError('Boolean value expected.')

    def strdeltautc(self, input_time_str):

        # 解析时间字符串为 datetime 对象
        input_time = datetime.strptime(input_time_str, "%H:%M:%S")
        # 减去 8 小时的时间间隔
        new_time = input_time - timedelta(hours=8)
        # 将结果转换为字符串
        new_time_str = new_time.strftime("%H:%M:%S")
        return new_time_str

    @MongoClientTradeGateway.error_handler
    def req_buy_order_insert(self, obj):
        current_time = datetime.now()
        formatted_time = current_time.strftime("%Y%m%d%H%M%S%f")[:-3]
        temp_filename = self.insert_order_path + '\\InPut_temp\\' + '1_buy.' + formatted_time + '.csv'
        with open(temp_filename, 'w', newline='') as file:
            writer = csv.writer(file)
            target_account_name = obj['accountName']
            acc = self.target_account_names_to_acc[target_account_name]
            account_id = self.account_id[acc]
            acct_type = self.acct_type[acc]

            oid = str(self.gen_order_id())[4:] #cats limit 16
            mid = str(obj['mid'])
            vol = int(obj["volume"])

            finacial_buy_s = obj["financingBuy"]
            side = '1'
            if finacial_buy_s == True:
                side = 'A'
            else:
                side = '1'
            tradingstyle = 2
            if not obj.get('tradingstyle') is None:
                tradingstyle = int(obj['tradingstyle'])
            participate = 0.1
            if not obj.get('participate') is None:
                participate = int(obj['participate'])

            order_vol = f'{vol:d}'
            int_stock_code = int(obj['ticker'])
            stock_code = get_exchange_from_ticker(int_stock_code)
            algo_type = decode_ordtype(obj['executionPlan']['order_type'])
            begin_time  = str(obj['executionPlan']['start_time']).replace(':','')
            end_time = str(obj['executionPlan']['end_time']).replace(':','')
            ord_param = f'AlgoType={algo_type};beginTime={begin_time};endTime={end_time}'

            if algo_type in ['VWAP', 'TWAP']:
                ord_param += f';limitPrice=0;participateRate=0;tradingStyle={tradingstyle}'
            elif algo_type in ['VWAP3', 'TWAP3', 'SmartTWAP3', 'SmartVWAP3']:
                ord_param += f';limitPrice=0;participateRate=0;tradingStyle={tradingstyle}'
            elif algo_type in ['KF_POV_Core']:
                ord_param += f';maxRate={participate}'
            #elif algo_type in ['KF_VWAP_Plus', 'KF_TWAP_Plus']:
            #s    ord_param += f';beginDate'

            insert_row = [oid, acct_type, account_id, stock_code, vol, side, "algo", ord_param]
            writer.writerow(insert_row)
            self.logger.info(f"[req_buy_order] insert_success (oid){oid} (mid){mid} (param){insert_row}")
        current_time = datetime.now()
        formatted_time = current_time.strftime("%Y%m%d%H%M%S%f")[:-3]
        final_filename = f"{self.insert_order_path}\\InPut\\1_buy.{formatted_time}.csv"
        shutil.move(temp_filename, final_filename)
        db_id = obj['_id']
        order_dict = {
            'oid' : oid,
            'mid' : mid,
            "db_id" : db_id,
            "ticker" : obj['ticker'],
            "side" : side,
            "volume" : order_vol,
            "accountName" : target_account_name,
            "order_type" : str(obj['executionPlan']["order_type"]),
            "start_time" : str(obj['executionPlan']["start_time"]),
            "end_time" : str(obj['executionPlan']["end_time"])
        }
        target_collection =  self.order_info_db[acc]['target']
        delete_query = {
                    '_id' : db_id
                }
        delete_res = target_collection.delete_one(delete_query)
        self.logger.info(f"[on_req_buy_order] delete (target){obj}")

        atx_order_collection = self.order_info_db[acc]['atx_order']
        local_id_list = []
        local_ref_list = []
        db_msg = {
                    "oid" : oid,
                    "mid" : mid,
                    "traded_vol" : 0,
                    "traded_amt" : 0,
                    "local_ids" : local_id_list,
                    "local_refs" : local_ref_list,
                    "order_msg": order_dict
                }
        query = {"oid": oid}
        res = atx_order_collection.replace_one(query, db_msg, True)
        self.logger.info(f"[on_req_order_insert] insert_cat_order_info (res){res} (msg){db_msg}")

        self.sids.append(mid)

        self.sid_to_req[mid] = order_dict
        self.oid_to_req[oid] = order_dict
        self.oid_to_mid[oid] = mid
        self.db_id_to_oid[db_id] = oid
        self.oid_to_acc[oid] = acc
        self.oid_to_traded[oid] = 0
        self.oid_to_traded_money[oid] = 0
        self.oid_to_local_ids[oid] = local_id_list
        self.oid_to_ref_list[oid] = local_ref_list

    @MongoClientTradeGateway.error_handler
    def req_sell_order_insert(self, obj):
        current_time = datetime.now()
        formatted_time = current_time.strftime("%Y%m%d%H%M%S%f")[:-3]
        temp_filename = self.insert_order_path + '\\InPut_temp\\' + '1_sell.' + formatted_time + '.csv'
        with open(temp_filename, 'w', newline='') as file:
            writer = csv.writer(file)
            target_account_name = obj['accountName']
            acc = self.target_account_names_to_acc[target_account_name]
            account_id = self.account_id[acc]
            acct_type = self.acct_type[acc]

            oid = str(self.gen_order_id())[4:] #cats limit 16
            mid = str(obj['mid'])
            vol =  int(obj["volume"])

            side = '2'
            tradingstyle = 2
            if not obj.get('tradingstyle') is None:
                tradingstyle = int(obj['tradingstyle'])
            participate = 10
            if not obj.get('participate') is None:
                participate = int(obj['participate'])
            order_vol = f'{vol:d}'
            int_stock_code = int(obj['ticker'])
            stock_code = get_exchange_from_ticker(int_stock_code)
            algo_type = decode_ordtype(obj['executionPlan']['order_type'])
            begin_time  = str(obj['executionPlan']['start_time']).replace(':','')
            end_time = str(obj['executionPlan']['end_time']).replace(':','')
            ord_param = f'AlgoType={algo_type};beginTime={begin_time};endTime={end_time}'

            if algo_type in ['VWAP', 'TWAP']:
                ord_param += f';limitPrice=0;participateRate=0;tradingStyle={tradingstyle}'
            elif algo_type in ['VWAP3', 'TWAP3', 'SmartTWAP3', 'SmartVWAP3']:
                ord_param += f';limitPrice=0;participateRate=0;tradingStyle={tradingstyle}'
            elif algo_type in ['KF_POV_Core']:
                ord_param += f';maxRate={participate}'

            insert_row = [oid, acct_type, account_id, stock_code, vol, side, "algo", ord_param]
            writer.writerow(insert_row)
            self.logger.info(f"[req_sell_order] insert_success (oid){oid} (mid){mid} (param){insert_row}")
        current_time = datetime.now()
        formatted_time = current_time.strftime("%Y%m%d%H%M%S%f")[:-3]
        final_filename = f"{self.insert_order_path}\\InPut\\1_sell.{formatted_time}.csv"
        shutil.move(temp_filename, final_filename)
        db_id = obj['_id']
        order_dict = {
            'oid' : oid,
            'mid' : mid,
            "db_id" : db_id,
            "ticker" : obj['ticker'],
            "side" : side,
            "volume" : vol,
            "accountName" : target_account_name,
            "order_type" : str(obj['executionPlan']["order_type"]),
            "start_time" : str(obj['executionPlan']["start_time"]),
            "end_time" : str(obj['executionPlan']["end_time"])
        }
        sell_target_collection =  self.order_info_db[acc]['sell_target']
        delete_query = {
                    '_id' : db_id
        }
        delete_res = sell_target_collection.delete_one(delete_query)
        self.logger.info(f"[on_req_sell_order] delete (target){obj}")
        atx_order_collection = self.order_info_db[acc]['atx_order']
        local_id_list = []
        local_ref_list = []
        db_msg = {
                    "oid" : oid,
                    "mid" : mid,
                    "traded_vol" : 0,
                    "traded_amt" : 0,
                    "local_ids" : local_id_list,
                    "local_refs" : local_ref_list,
                    "order_msg": order_dict
                }
        query = {"oid": oid}
        res = atx_order_collection.replace_one(query, db_msg, True)
        self.logger.info(f"[on_req_order_insert] insert_cat_order_info (res){res} (msg){db_msg}")

        self.sids.append(mid)
        self.sid_to_req[mid] = order_dict
        self.oid_to_req[oid] = order_dict
        self.oid_to_mid[oid] = mid
        self.db_id_to_oid[db_id] = oid
        self.oid_to_acc[oid] = acc
        self.oid_to_traded[oid] = 0
        self.oid_to_traded_money[oid] = 0
        self.oid_to_local_ids[oid] = local_id_list
        self.oid_to_ref_list[oid] = local_ref_list
    #使用的是执行统计输出，统计的是算法母单
    @MongoClientTradeGateway.error_handler
    def monitor_algo_order_update(self):
        print ("[monitor_algo_order_update]")
        while not self.is_stopped:
            order_filename = self.recv_msg_dir + '\\' + FILES.ALGO_ORDERS + '.dbf'
            rtn_order_table = dbf.Table(filename=order_filename,codepage='utf8', on_disk=True)
            with rtn_order_table.open(mode=dbf.READ_WRITE):
                algo_index = self.algo_index
                if algo_index == len(rtn_order_table) - 1 or len(rtn_order_table) == 0:
                    self.logger.info(f"[monitor_order_update] not_change (len){len(rtn_order_table)}")
                    time.sleep(self.order_scan_interval)
                    continue
                while rtn_order_table[algo_index+1] is not rtn_order_table.last_record:
                    algo_index += 1
                    record = rtn_order_table[algo_index]
                    self.update_algo_order(record)
                algo_index += 1
                self.update_algo_order(rtn_order_table.last_record)
            time.sleep(self.order_scan_interval)
    @MongoClientTradeGateway.error_handler
    def update_algo_order(self,record):
        oid = str(record.ID).strip(' ')
        if oid == "":
            self.logger.warning(f"[update_algo_order] no_oid (record){record}")
            return
        acc = ""
        mid = ""
        order_dict = {}
        algo_id = str(record.INSTANCEID).strip(' ')
        #order_no 返回用
        self.algoid_to_oid[algo_id] = oid
        self.oid_to_algoid[oid] = algo_id
        if not oid in self.oid_to_acc: #shut_down or other problem
            trade_acc = str(record.ACCT).strip(' ')
            if trade_acc not in self.account_id_to_acc:
                self.logger.error(f"[update_algo_order] can't_parse_trade_acc {trade_acc}")
                return
            acc = self.account_id_to_acc[trade_acc]
            order_info_collection = self.order_info_db[acc]['atx_order'] 
            query = {'oid' : oid}
            order_info_target = order_info_collection.find_one(query)
            if not order_info_target is None:
                order_dict = order_info_target['order_msg']
                mid = order_info_target['mid']
                ref_list = order_info_target['local_refs']
                self.oid_to_traded_money[oid] = 0
                self.oid_to_traded[oid] = 0
                self.oid_to_mid[oid] = mid
                self.oid_to_req[oid] = order_dict
                self.oid_to_ref_list[oid] = ref_list
            else:
                self.logger.error(f"[update_algo_order] can't_find_oid (oid){oid}")
                return
        else:
            acc = self.oid_to_acc[oid]
            order_dict = self.oid_to_req[oid]
            mid = self.oid_to_mid[oid]
        atx_order_msg = {
                    "algo_id" : algo_id
                }
        query = {"oid": oid}
        self.update_atx_order(acc, query, atx_order_msg)
        ticker = record.Symbol.split('.')[0]
        exchange = decode_exchange_id(str(record.Symbol.split('.')[1]).strip(' '))
        volume = int(order_dict['volume'])
        price = float(str(record.FILLPRICE).strip(' '))

        tg_name = order_dict['accountName'].split('@')[0]
        account_name = order_dict['accountName'].split('@')[1]
        order_type = order_dict['order_type']
        start_time = order_dict['start_time']
        end_time = order_dict['end_time']

        target_type = side_to_target_type(str(order_dict['side']))
        filled_vol = int(str(record.QTY_FILLED).strip(' '))

        status = decode_cats_status(int(str(record.STATUS).strip(' ')), volume, filled_vol)
        replace_collection = self.tradelog_db[acc]['order']
        query = {'mid':mid, '_id': int(oid)}
        replace_target = replace_collection.find_one(query)
        if not replace_target is None:
            utc_update_time = replace_target['dbTime']
            order_msg = {
                    "_id": int(oid),
                    "tg_name": tg_name,  # 对应 account_info._id
                    "exchange": exchange, # 对应 tlclient.trader.constant 包 ExchangeID
                    "target_type": target_type,    # 'buy' | 'sell' | 'limit_sell'
                    "volume": volume,  # 订单的volume
                    "price": price,  # 实际成交均价
                    "order_type": order_type,  # algotype:200,目前没有其他的
                    "ticker": ticker,
                    "mid": mid,  # target中对应的 母单sid
                    "accountName": account_name,  # 对应 account_info.account_name
                    "algo_args": {  # 具体的算法参数
                        "order_type": order_type,  # 此柜台不使用下单时要求的算法
                        "start_time": start_time,
                        "end_time": end_time
                    },
                    "status": status,  # 'active' | 'filled' | 'canceled'
                    "filled_vol": filled_vol,  # 实际成交的 volume
                    "dbTime": utc_update_time,
            }

            res = replace_collection.replace_one(query, order_msg, True)
            self.logger.info(f"[rtn_order] (res){res} (order_msg){order_msg}")
        else:
            str_update_time = str(str(record.WRITE_TIME).strip(' ').split('@')[1])[:-4] #去除毫秒部分
            update_time = datetime.strptime(str_update_time, "%Y-%m-%d %H:%M:%S")
            utc_update_time = update_time - timedelta(hours=8)
            #utc_update_time = replace_target['dbTime']
            order_msg = {
                    "_id": int(oid),
                    "exchange": exchange,
                    "target_type": target_type,    # 'buy' | 'sell' | 'limit_sell'
                    "volume": volume,  # 订单的volume
                    "price": price,  # 实际成交均价
                    "order_type": order_type,  # algotype:200,目前没有其他的
                    "ticker": ticker,
                    "mid": mid,  # target中对应的 母单sid
                    "accountName": account_name,  # 对应 account_info.account_name
                    "algo_args": {  # 具体的算法参数
                        "order_type": order_type,  # 此柜台不使用下单时要求的算法
                        "start_time": start_time,
                        "end_time": end_time
                    },
                    "status": status,  # 'active' | 'filled' | 'canceled'
                    "filled_vol": filled_vol,  # 实际成交的 volume
                    "dbTime": utc_update_time,
            }
            res = replace_collection.replace_one(query, order_msg, True)
            self.logger.info(f"[rtn_order] (res){res} (order_msg){order_msg}")
    #使用的是执行统计输出，统计的是算法母单
    #0 已报 #2 filled 全成 #4 canceled #8 rejected
    @MongoClientTradeGateway.error_handler
    def monitor_order_update(self):
        print ("[monitor_order_update]")
        while not self.is_stopped:

            order_filename = self.recv_msg_trade_dir + '\\' + FILES.ORDERS + '.dbf'
            rtn_order_table = dbf.Table(filename=order_filename,codepage='utf8', on_disk=True)
            rtn_order_table.open(mode=dbf.READ_WRITE)
            order_index = self.order_index
            if order_index == len(rtn_order_table) - 1 or len(rtn_order_table) == 0:
                self.logger.info(f"[monitor_order_update] not_change (len){len(rtn_order_table)}")
                time.sleep(self.order_scan_interval)
                continue

            while rtn_order_table[order_index+1] is not rtn_order_table.last_record:
                order_index += 1
                record = rtn_order_table[order_index]
                order_status = int(str(record.ORD_STATUS).strip(' '))
                if order_status in [5, 6]:
                    self.logger.warning(f"[monitor_order_update] order_rejected! (msg){str(record.ERR_MSG).strip(' ')}")        #self.logger.warn(f"[monitor_order_update] order_rejected!")
                elif order_status in [0, 1, 2, 3, 4]:
                    self.update_order(record)
                else:
                    self.logger.info("[monitor_order_update] order_not_finished_or_not_init")

            order_index += 1
            record = rtn_order_table[order_index]
            order_status = int(str(record.ORD_STATUS).strip(' '))
            if order_status in [5, 6]:
                self.logger.warning(f"[monitor_order_update] order_rejected! (msg){str(record.ERR_MSG).strip(' ')}")
            elif order_status in [0, 1, 2, 3, 4]:
                self.update_order(record)
            rtn_order_table.close()
            time.sleep(self.order_scan_interval)
    @MongoClientTradeGateway.error_handler
    def update_order(self,record):
        algo_id = str(record.CORR_ID).strip(' ')
        if algo_id not in self.algoid_to_oid:
            self.logger.warning(f'[update_order] can not match algo_id {algo_id}')
            return
        oid = self.algoid_to_oid[algo_id]
        acc = ""
        mid = ""
        order_dict = {}
        mudan_id = str(record.ORD_NO).strip(' ') #orderno 返回用
        if not oid in self.oid_to_acc: #shut_down or other problem
            trade_acc = str(record.ACCT).strip(' ')
            if trade_acc not in self.account_id_to_acc:
                self.logger.error(f"[update_order] can't_parse_trade_acc {trade_acc}")
                return
            acc = self.account_id_to_acc[trade_acc]
            order_info_collection = self.order_info_db[acc]['atx_order']
            query = {'oid' : oid}
            order_info_target = order_info_collection.find_one(query)
            if not order_info_target is None:
                order_dict = order_info_target['order_msg']
                mid = order_info_target['mid']
                ref_list = order_info_target['local_refs']
                self.oid_to_traded_money[oid] = 0
                self.oid_to_traded[oid] = 0
                self.oid_to_mid[oid] = mid
                self.oid_to_req[oid] = order_dict
                self.oid_to_ref_list[oid] = ref_list
                self.oid_to_algoid[oid] = algo_id
            else:
                self.logger.error(f"[update_order] can't_find_oid (oid){oid}")
                return
        else:
            acc = self.oid_to_acc[oid]
            order_dict = self.oid_to_req[oid]
            mid = self.oid_to_mid[oid]
        self.ref_to_oid[mudan_id] = oid
        if mudan_id not in self.oid_to_ref_list[oid]:
            self.oid_to_ref_list[oid].append(mudan_id)
        atx_query = {'oid' : oid}
        update_msg = {
            "local_refs" : self.oid_to_ref_list[oid]
        }
        self.update_atx_order(acc, atx_query, update_msg)

    @MongoClientTradeGateway.error_handler
    def monitor_algo_order_insert(self):
        while not self.is_stopped:
            print ("[monitor_algo_order_insert]")
            for acc in self.accounts_run:
                target_account_name = self.target_account_names[acc]
                buy_query = {"accountName": target_account_name}
                buy_targets = self.order_info_db[acc]["target"].find(buy_query)
                if buy_targets.count() == 0:
                    self.logger.warning(f"[monitor_buy_order] no_buy_target (acc){acc}")
                    continue
                for target in buy_targets:
                    if target['_id'] not in self.order_db_ids:
                        self.order_db_ids.append(target['_id'])
                        self.req_buy_order_insert(target)
                    else:
                        self.logger.warning(f"[monitor_algo_order_insert] _id_existed (_id){target['_id']}")
            for acc in self.accounts_run:
                target_account_name = self.target_account_names[acc]
                sell_query = {"accountName": target_account_name}
                sell_targets = self.order_info_db[acc]["sell_target"].find(sell_query)
                if sell_targets.count() == 0:
                    self.logger.warning(
                        f"[monitor_sell_order] no_sell_target (acc){acc}")
                    continue
                for sell_target in sell_targets:
                    if sell_target['_id'] not in self.sell_order_db_ids:
                        self.sell_order_db_ids.append(sell_target['_id'])
                        self.req_sell_order_insert(sell_target)
                    else:
                        self.logger.warning(f"[monitor_algo_order_insert] _id_existed (_id){sell_target['_id']}")
            time.sleep(self.scan_interval)
    #得到的数据 类似于子单的rtn_order
    @MongoClientTradeGateway.error_handler
    def monitor_trade_update(self):
        while not self.is_stopped:
            print("[monitor_trade_update]")
            trade_filename = self.recv_msg_trade_dir + '\\' + FILES.TRADES + '.dbf'
            rtn_trade_table = dbf.Table(filename=trade_filename,codepage='utf8', on_disk=True)
            with rtn_trade_table.open(mode=dbf.READ_WRITE):
                trade_index = self.trade_index
                if trade_index == len(rtn_trade_table)-1 or len(rtn_trade_table) == 0:
                    self.logger.info(f"[monitor_trade_update] not_change (len){len(rtn_trade_table)}")
                    time.sleep(self.trade_scan_interval)
                    continue
                else:
                    while rtn_trade_table[trade_index +1] is not rtn_trade_table.last_record:
                        trade_index += 1
                        record = rtn_trade_table[trade_index]
                        self.on_trade_msg(record)
                    trade_index += 1
                    record = rtn_trade_table[trade_index]
                    self.on_trade_msg(record)

            time.sleep(self.trade_scan_interval)

    @MongoClientTradeGateway.error_handler
    def on_trade_msg(self, record):
        fill_qty = int(str(record.FILLQTY).strip(' '))
        fill_price = float(str(record.FILLPRICE).strip(' '))
        if fill_qty == 0 or fill_price == 0:
            self.logger.warning(f"[on_trade_msg] nothing_traded")
            return
        mudan_id = str(record.ORDERNO).strip(' ')
        if mudan_id == "":
            self.logger.warning(f"[on_trade_msg] no_mid (mudan_id){mudan_id}")
            return
        if mudan_id not in self.ref_to_oid:
            self.logger.warning(f"[on_trade_msg] no_oid (mudan_id){mudan_id}")
            return
        oid = self.ref_to_oid[mudan_id]
        acc = ""
        mid = ""
        order_dict = {}
        if not oid in self.oid_to_acc: #shut_down or other problem
            trade_acc = str(record.ACCOUNT).strip(' ')
            acc = self.account_id_to_acc[trade_acc]
            order_info_collection = self.order_info_db[acc]['atx_order']
            query = {'oid' : oid}
            order_info_target = order_info_collection.find_one(query)

            if not order_info_target is None:
                order_dict = order_info_target['order_msg']
                mid = order_info_target['mid']
                ref_list = order_info_target['local_refs']
                self.oid_to_local_ids[oid] = order_info_target['local_ids']
                self.oid_to_traded[oid] = 0
                self.oid_to_traded_money[oid] = 0
                self.oid_to_mid[oid] = mid
                self.oid_to_req[oid] = order_dict
                self.oid_to_ref_list[oid] = ref_list
            else:
                self.logger.error(f"[on_trade_msg] can't_find_oid (oid){oid}")
                return
        else:
            acc = self.oid_to_acc[oid]
            order_dict = self.oid_to_req[oid]
            mid = self.oid_to_mid[oid]
        _id = str(record.TRADENO).strip(' ')
        if _id in self.oid_to_local_ids[oid]:
            #self.logger.warning(f"[on_trade_msg] duplicate_trade_msg (record){record}")
            return
        else:
            self.oid_to_local_ids[oid].append(_id)#外部委托编号，出现在这个位置主要是凑数
        ticker = str(record.Symbol).split('.')[0]
        exchange = decode_exchange_id(str(record.Symbol).split('.')[1].strip(' '))
        traded_vol = fill_qty
        traded_price = fill_price
        trade_amt = float(traded_price * traded_vol)
        order_type = str(record.TRDTYPE).strip(' ')
        target_type = decode_cats_side(str(record.TRADSIDE).strip(' '))
        if target_type == 1 or target_type == 3:
            self.oid_to_traded_money[oid] += trade_amt
            self.oid_to_traded[oid] += traded_vol
        elif target_type == 2:
            self.oid_to_traded_money[oid] -= trade_amt
            self.oid_to_traded[oid] -= traded_vol
        entrust_vol = int(order_dict['volume'])
        entrust_price = 0 #没有
        dbTime = datetime.now(timezone.utc)
        str_trade_time = str(record.TIME).strip(' ')[:-4]
        trade_time = datetime.strptime(str_trade_time, "%Y-%m-%d %H:%M:%S")
        utc_trade_time = trade_time - timedelta(hours=8)
        target_account_name = order_dict['accountName']
        if target_account_name not in self.target_account_names_to_acc: #切换账号了，缓存无法使用
            return
        acc = self.target_account_names_to_acc[target_account_name]
        log_account_name = order_dict['accountName'].split('@')[1]
        tg_name =  order_dict['accountName'].split('@')[0]
        trade_collection = self.tradelog_db[acc]['tg_trade']
        replace_trade_query = { "trade_ref": _id, "oid" : int(oid), "mid": mid, "accountName": log_account_name}
        trade_target = trade_collection.find_one(replace_trade_query)
        if not trade_target is None:
            dbTime = trade_target['dbTime']
        db_msg = {
                    "trade_ref": _id, # 用的是非凸的子单order_id 经测试之前用的trade_ref 经常为空
                    "oid": int(oid),
                    "tg_name": tg_name,
                    "exchange": exchange,
                    "ticker": ticker,
                    "traded_vol": traded_vol,
                    "traded_price": traded_price,
                    "order_type": order_type,
                    "side": target_type,  # 对应 tlclient.trader.constant 包 Side
                    "entrust_vol": entrust_vol,
                    "entrust_price": entrust_price,
                    "dbTime": dbTime,
                    "mid": mid,  # 对应订单中的 mid
                    "commission": 0,  # 没有
                    "trade_time": utc_trade_time,  # 具体交易时间
                    "accountName":  log_account_name,  # 对应 account_info.account_name
                }

        db_res = trade_collection.replace_one(replace_trade_query, db_msg, True)
        self.logger.info(
            f"[rtn_trade] (db_res){db_res} (db_msg){db_msg} (traded_vol){traded_vol} (traded_price){traded_price}")

        atx_order_collection = self.order_info_db[acc]['atx_order']
        atx_order_msg = {
                    "local_ids" : self.oid_to_local_ids[oid],
                }
        query = {"oid": oid}
        res = self.update_atx_order(acc, query, atx_order_msg)
        self.logger.info(f"[on_trade_msg] update_cat_order_info (res){res} (msg){atx_order_msg}")
    @MongoClientTradeGateway.error_handler
    def monitor_dbf_pos(self):
        print ("[monitor_dbf_pos]")
        while not self.is_stopped:
            order_filename = self.recv_msg_dir + '\\' + FILES.POSITIONS  + '.dbf'
            position_table = dbf.Table(filename=order_filename,codepage='utf8', on_disk=True)
            for acc in self.accounts_run:
                tg_position_collection = self.order_info_db[acc]['tg_equity_position']
                remove = tg_position_collection.delete_many({'account_name': self.log_account_names[acc], 'tg_name': self.target_account_names[acc]})
                self.logger.info(f"[monitor_dbf_pos] delete_old_position_info (remove){remove} ")
            with position_table.open(mode=dbf.READ_WRITE):
                pos_index = -1
                while position_table[pos_index+1] is not position_table.last_record:
                    pos_index += 1
                    record = position_table[pos_index]
                    self.update_dbf_position(record)

                pos_index += 1
                self.update_dbf_position(position_table.last_record)
            time.sleep(self.pos_interval)

    def update_dbf_position(self, record):
        a_type = str(record.A_TYPE).strip(' ')
        if a_type not in ['P']: #其他类型
            return
        trade_acc = str(record.ACCT).strip(' ')

        if trade_acc not in self.account_id_to_acc:
            self.logger.warning(f"[update_dbf_position] can't_parse_trade_acc {trade_acc}")
            return
        acc = self.account_id_to_acc[trade_acc]
        account_name = self.log_account_names[acc]
        tg_name = self.target_account_names[acc]
        ticker = record.S1.split('.')[0].strip(' ')
        exchange = decode_exchange_id(str(record.S1.split('.')[1]).strip(' '))
        td_pos = int(str(record.S2).strip(' '))
        yd_pos = int(str(record.S3).strip(' '))
        #new_id = tg_name + '@' + account_name + '@' + ticker
        pos_collection = self.order_info_db[acc]['tg_equity_position']
        query = {'tg_name':tg_name, 'account_name': account_name, "ticker": ticker}
        pos_msg = {
                "account_name": account_name,
                "tg_name": tg_name,
                "ticker": ticker,
                "exchange": exchange,
                "direction": "long", # long/short，没有short部分就只存long部分
                "avail_pos": yd_pos, # 昨仓
                "total_pos": td_pos, # 今仓，TODO 需要统计下不同券商盘中，对于卖出的position 是直接从 yd_pos 上减，还是在 td_pos 增加一个负的值。
                "cost": 0, #没有
                "type": "stock",
                "updated_at": datetime.utcnow()
              }
        res = pos_collection.replace_one(query, pos_msg, True)
        self.logger.info(f"[update_dbf_position] (res){res.modified_count} (order_msg){pos_msg}")
    def start(self):
        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGINT, self.signal_handler)
        _msg = f"[atx_login] {self.log_name} start (time){datetime.now()}"
        self.send_to_user(logger=self.logger, url_list=self.url_list, msg=_msg)
        self.monitor()

    def close(self):
        self.position_order_table.close()
        return super().close()
    
    def join(self):
        while self.is_stopped == False:
            time.sleep(0.01)
            if self.is_stopped:
                self.logger.info(
                    "[close] main thread is stopped,active_orders message will lose")
                self.close()

if __name__ == "__main__":

    sys.stdout.reconfigure(encoding='utf-8')
    description = "cat_uni_server,get target from mongodb and serve cat"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('-e' , '--end_time', dest='end_time', default='15:30')
    _config_filename = "C:/Users/Administrator/Desktop/cat_uni_batandconfig/cat_uni_server_config.json"
    parser.add_argument('-p', '--config_filepath', dest= 'config_filepath', default= _config_filename)

    args = parser.parse_args()
    print (f"(args){args}")

    td = CatServer(args.config_filepath, args.end_time)
    td.start()

    td.join()


import sys
import os
from datetime import datetime

def decode_exchange_id(suffix):
    exchange = 0
    if suffix == 'SH':
        exchange = 101        #EXCHANGE_SSE 
    elif suffix == 'SZ':
        exchange = 102        #EXCHANGE_SZE 
    
    return exchange

def get_log_default_path():
    # python2: linux2, python3: linux
    if sys.platform.startswith("linux") or sys.platform == "darwin":
        dirs = "/shared/log"
    elif sys.platform == "win32":
        dirs = os.path.join(get_windows_first_disk() + "/tmp/linker/log")
    else:
        dirs = '.'

    return dirs


def get_digit_from_env(env_name, default_num):
    num = str(os.environ.get(env_name))
    return int(num) if num.isdigit() else default_num

def get_log_given_path(path):
    dirs = os.path.join(path)
    return path

def get_exchange_from_ticker(ticker:int):
    stock_code = f'{ticker:06d}'
    if stock_code.startswith('00'):
        stock_code += '.SZ'
    elif stock_code.startswith('6'):
        stock_code += '.SH'
    elif stock_code.startswith('5'):
        stock_code += '.SH'
    else:
        stock_code += '.SZ'
    return stock_code

def check_today_index(input_datetime_str):
    input_datetime = datetime.strptime(input_datetime_str, "%Y-%m-%d %H:%M:%S")

    # 获取当前日期的 datetime 对象
    current_datetime = datetime.now()

    # 判断是否在今天
    if input_datetime.date() == current_datetime.date():
        return True

    return False


from datetime import datetime, timezone
import os
import queue
import threading
import time
import json
import requests
import signal
import sys
import traceback
from functools import wraps
from concurrent.futures import ThreadPoolExecutor
import dbf
import argparse

try:
    import thread
except ImportError:
    import _thread as thread

from pymongo import MongoClient, ASCENDING, DESCENDING
from .logger import Logger
from .utils import decode_exchange_id
from .constants import DbfFiles

LL9 = 1000000000

class MongoClientTradeGateway(object):
    def __init__(self, config_filename, endtime, gateway_name):
        
        self.load_gateway_setting(config_filename)
        self.logger = Logger.get_logger(self.log_name, self.log_file_path)
        self.gen_local_id()
        self.endtime = endtime
        self.is_stopped = False
        self.start_mongodb()
        
        self.thread_pool = ThreadPoolExecutor(10)
        self.sell_orderlock = threading.Lock()
        self.cancel_orderlock = threading.Lock()
        
        self.order_db_ids = []
        self.sell_order_db_ids = []
        self.cancel_order_ids = []
        self.gateway_name = gateway_name
        
        self.date = self.get_date_today()
        
        
    def error_handler(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                return func(self,*args, **kwargs)
            except Exception as e:
                err = traceback.format_exc()
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback, limit=None, file=sys.stdout)
                self.logger.error(f'[{wrapper}] (exception){err}')
                self.send_error_to_user(self.logger, self.error_url_list, err)
                return 'error'
        return wrapper
        
    def load_gateway_setting(self, config_filename):
        try:
            #固定配置文件的名字
            #config_filename = os.path.join(config_filename, 'atx_cicc_server_config.json')
            #f = open(config_filename, encoding="utf-8")
            f = open(config_filename, encoding='gbk')
            setting = json.load(f)
            
            log_path = setting['log_filepath']
            self.log_file_path = log_path.replace('/', '\\')
            self.url_list = setting.get('url_list')
            self.error_url_list = setting.get('error_url_list')

            self.log_name = setting['log_name']
            self.scan_interval = setting['scan_interval']
            self.order_scan_interval = setting['order_scan_interval']
            self.trade_scan_interval = setting['trade_scan_interval']
            self.pos_interval = setting['pos_interval']
            self.acc_interval = setting['acc_interval']
            
            self.accounts_config = setting['accounts']
            self.accounts_run = setting['run']
            
            self.config = {}
            self.account_id = {}
            self.account_id_to_acc = {}
            self.product_names = {}
            self.log_account_names = {}
            self.tgnames = {}
            self.mongo_host = {}
            self.mongo_port = {}
            self.tradingaccount_user = {}
            self.tradingaccount_pwd = {}
            self.tradinglog_user = {}
            self.tradinglog_pwd = {}
            self.target_account_names = {}
            self.target_account_names_to_acc = {}
            self.contract_type = {}
            for acc in self.accounts_run:
                self.config[acc] = setting['accounts'][acc]
                config = self.config[acc]
                self.account_id[acc] = config['account_id']
                self.account_id_to_acc[config['account_id']] = acc
                self.product_names[acc] = config['product_name']
                self.log_account_names[acc] = config['account_name']
                self.tgnames[acc] = config['equity_tg_name']
                self.target_account_names[acc] = config['equity_tg_name'] + "@" + config['account_name']
                self.target_account_names_to_acc[self.target_account_names[acc]] = acc
                self.mongo_host[acc] = config['mongoHost']
                self.mongo_port[acc] = config['mongoPort']
                datadbuser = config['databaseUser']
                self.tradingaccount_user[acc] = datadbuser['tradingAccount']['user']
                self.tradingaccount_pwd[acc] = datadbuser['tradingAccount']['password']
                self.tradinglog_user[acc] = datadbuser['tradingLog']['user']
                self.tradinglog_pwd[acc] = datadbuser['tradingLog']['password']
            
        except Exception as e:
            err = traceback.format_exc()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback, limit=None, file=sys.stdout)
            print(f"load config failed! (exception){err}")
            exit(0)
    
    def start_mongodb(self):
        try:
            self.db_client = {}
            self.order_info_db = {}
            self.tradelog_db = {}
            for acc in self.accounts_run:
                self.db_client[acc] = MongoClient(
                    self.mongo_host[acc], self.mongo_port[acc], connectTimeoutMS=10000)
                db_client = self.db_client[acc]
                if self.tradingaccount_user[acc] != '' and self.tradingaccount_pwd[acc] != '':
                    db_client["tradingAccount"].authenticate(
                        self.tradingaccount_user[acc], self.tradingaccount_pwd[acc], mechanism='SCRAM-SHA-1')
                self.order_info_db[acc] = db_client["tradingAccount"]
                
                if self.tradinglog_user[acc] != '' and self.tradinglog_pwd[acc] != '':
                    db_client["tradingLog"].authenticate(
                        self.tradinglog_user[acc], self.tradinglog_pwd[acc], mechanism='SCRAM-SHA-1')
                db_client.server_info()
                self.tradelog_db[acc] = db_client["tradingLog"] 
                
        except Exception as e:
            err = traceback.format_exc()
            self.send_error_to_user(self.error_url_list, err)
            self.logger.error(f'[init] DB_connect_failed! (exception){err}')
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback, limit=None, file=sys.stdout)
            exit()

    #通过tradingAccount.accountinfo表获取产品属性
    def get_account_info(self):
        for acc in self.accounts_run:
            product_name = self.product_names[acc]
            query = {"product_name": product_name}
            account_info_collection = self.order_info_db[acc]['account_info']
            account_info = account_info_collection.find_one(query)
            if account_info == None:
                self.logger.error(
                    f"[get_account_info] can't_find_account_info (product_name){product_name}")
                continue
            tgname = account_info['equity_tg_name']
            self.tgnames[acc] = tgname
            log_account_name = account_info['account_name']
            self.log_account_names[acc] = log_account_name
            target_account_name = tgname + '@' + log_account_name
            self.target_account_names[acc] = target_account_name # 下单时用self
            self.logger.info(
                f"[get_account_info] (tg_name){self.tgnames} (logacc_name){self.log_account_names} (target_accountnames){self.target_account_names}") 

    def send_to_user(self, logger, url_list, msg):
        try:
            if len(url_list) == 0:
                logger.info("[send_to_user] send_message_failed")
                return

            payload_message_feishu = {
                "msg_type": "text",
                "content": {
                    "text": msg
                }
            }

            headers = {
                'Content-Type': 'application/json'
            }

            payload_message_dingding = {
                "msgtype": "text",
                "text": {"content": msg},
                "at": {
                    "atMobiles": [""],
                    "isAtAll": "false"  # @所有人 时为true，上面的atMobiles就失效了
                }
            }
            response = requests.request("POST", url_list[0], headers=headers, data=json.dumps(payload_message_feishu))
            data = response.json()
            logger.info(f"[send_to_user] (response){response} (data){data}")
            if len(url_list) > 1:
                response_ding = requests.request("POST", url_list[1], headers=headers, data=json.dumps(payload_message_dingding))
                data_ding = response_ding.json()
                logger.info(f"[send_to_user] (response){response_ding} (data){data_ding}")
        except Exception as e:
            err = traceback.format_exc()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback, limit=None, file=sys.stdout)
            logger.error(f'[send_to_user] send_message_failed(exception){err}')
            
    def send_error_to_user(self, logger, url_list, msg):
        try:
            if len(url_list) == 0:
                logger.info("[send_error_to_user] send_message_failed")
                return

            payload_message_feishu = {
                "msg_type": "text",
                "content": {
                    "text": msg
                }
            }

            headers = {
                'Content-Type': 'application/json'
            }

            payload_message_dingding = {
                "msgtype": "text",
                "text": {"content": msg},
                "at": {
                    "atMobiles": [""],
                    "isAtAll": "false"  # @所有人 时为true，上面的atMobiles就失效了
                }
            }
            response = requests.request("POST", url_list[0], headers=headers, data=json.dumps(payload_message_feishu))
            data = response.json()
            logger.info(f"[send_error_to_user] (response){response} (data){data}")
            if len(url_list) > 1:
                response_ding = requests.request("POST", url_list[1], headers=headers, data=json.dumps(payload_message_dingding))
                data_ding = response_ding.json()
                logger.info(f"[send_error_to_user] (response){response_ding} (data){data_ding}")
        except Exception as e:
            err = traceback.format_exc()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback, limit=None, file=sys.stdout)
            logger.error(f'[send_error_to_user] send_message_failed(exception){err}')

    def signal_handler(self, signum=None, frame=None):
        self.is_stopped = True

    def gen_local_id(self):
        self.id_base = 1377779200 * LL9
        self.sp = time.time_ns()
        self.local_id = self.sp - self.id_base

    def gen_order_id(self):
        self.local_id += 1
        return self.local_id

    def get_date_today(self):
        dt = datetime.now()
        date = str(dt.strftime("%Y%m%d"))

        self.logger.info(f"[{self.gateway_name}_get_date_today] (date){date}")

        return date

    @error_handler
    def monitor_dbf_asset(self):
        print ("[monitor_dbf_asset]")
        while not self.is_stopped:
            order_filename = self.recv_msg_dir + '\\' + DbfFiles.ACCOUNTS + self.date + '.dbf'
            asset_table = dbf.Table(filename=order_filename,codepage='utf8', on_disk=True)
            with asset_table.open(mode=dbf.READ_WRITE):
                asset_index = -1
                while asset_table[asset_index+1] is not asset_table.last_record:
                    asset_index += 1
                    record = asset_table[asset_index]
                    self.update_dbf_asset(record)

                asset_index += 1
                self.update_dbf_asset(asset_table.last_record)
            time.sleep(self.acc_interval)

    def update_dbf_asset(self, record):
        trade_acc = str(record.ClientName).strip(' ')
        en_balance = record.EnBalance
        market_amt = record.MarketAmt

        if trade_acc not in self.account_id_to_acc:
            self.logger.warning(f"[update_dbf_asset] can't_parse_trade_acc {trade_acc}")
            return
        acc = self.account_id_to_acc[trade_acc]
        tg_name = self.tgnames[acc]
        account_name = self.log_account_names[acc]

        asset_collection = self.order_info_db[acc]['TestEquityAccount']
        query = {'_id':tg_name, 'accountName': account_name}
        asset_msg = {
            "_id": tg_name, # 对应之前的 TG 名称
            "accountName": account_name,  # 产品名称
            "avail_amt": en_balance,  # 可用资金（可现金买入的资金）
            "balance": en_balance + market_amt, # 净资产.  最好是由券商那边提供，通常普通账户是 可用资金(包含预扣费用) + 市值；信用账户是总资金 - 总负债
            "holding": market_amt, # 市值,
        }
        res = asset_collection.replace_one(query, asset_msg, True)
        self.logger.info(f"[update_dbf_asset] (res){res} (order_msg){asset_msg}")

    @error_handler
    def monitor_dbf_pos(self):
        print ("[monitor_dbf_pos]")
        while not self.is_stopped:
            order_filename = self.recv_msg_dir + '\\' + DbfFiles.POSITIONS + self.date + '.dbf'
            position_table = dbf.Table(filename=order_filename,codepage='utf8', on_disk=True)
            for acc in self.accounts_run:
                tg_position_collection = self.order_info_db[acc]['tg_equity_position']
                remove = tg_position_collection.delete_many({'account_name': self.log_account_names[acc], 'tg_name': self.target_account_names[acc]})
                self.logger.info(f"[monitor_dbf_pos] delete_old_position_info (remove){remove} ")
            with position_table.open(mode=dbf.READ_WRITE):
                pos_index = -1
                while position_table[pos_index+1] is not position_table.last_record:
                    pos_index += 1
                    record = position_table[pos_index]
                    self.update_dbf_position(record)

                pos_index += 1
                self.update_dbf_position(position_table.last_record)
            time.sleep(self.pos_interval)
        
        
    def update_dbf_position(self, record):
        trade_acc = str(record.ClientName).strip(' ')

        if trade_acc not in self.account_id_to_acc:
            self.logger.warning(f"[update_dbf_position] can't_parse_trade_acc {trade_acc}")
            return
        acc = self.account_id_to_acc[trade_acc]
        
        account_name = self.log_account_names[acc]
        tg_name = self.target_account_names[acc]
        ticker = record.Symbol.split('.')[0].strip(' ')
        exchange = decode_exchange_id(str(record.Symbol.split('.')[1]).strip(' '))
        td_pos = record.CurrentQ
        yd_pos = record.EnableQty
        #new_id = tg_name + '@' + account_name + '@' + ticker
        pos_collection = self.order_info_db[acc]['tg_equity_position']
        query = {'tg_name':tg_name, 'account_name': account_name, "ticker": ticker}
        pos_msg = {
              "account_name": account_name,
              "tg_name": tg_name,
              "ticker": ticker,
              "exchange": exchange,
              "direction": "long", # long/short，没有short部分就只存long部分
              "avail_pos": yd_pos, # 昨仓
              "total_pos": td_pos, # 今仓，TODO 需要统计下不同券商盘中，对于卖出的position 是直接从 yd_pos 上减，还是在 td_pos 增加一个负的值。
              "cost": 0, #没有
              "type": "stock",
              "updated_at": datetime.utcnow()
              
                        }
        update_msg = {'$set' : pos_msg}
        res = pos_collection.replace_one(query, pos_msg, True)
        self.logger.info(f"[update_dbf_position] (res){res.modified_count} (order_msg){pos_msg}")

    @error_handler
    def date_change(self):
        while not self.is_stopped:
                time_now = datetime.now()
                _dt_endtime = datetime.strptime(self.endtime, "%H:%M")
                dt_endtime = datetime.combine(time_now, _dt_endtime.time())
                if time_now > dt_endtime:
                    self.close()
                else:
                    self.logger.info(f"[{self.gateway_name}_date_change] not_closed (now){time_now}")

                time.sleep(60)

    def close(self):
        self.is_stopped = True

        msg = f"[{self.gateway_name}_close] (close_time){datetime.now()}"

        self.send_to_user(self.logger, self.url_list, msg)
        os._exit(0)
    
    def update_position_date_open(self):
        for acc in self.accounts_run:
            collection = self.order_info_db[acc]['EquityPosition']
            target_account_name = self.target_account_names[acc]
            targets_query = {'accountName' : target_account_name}
            targets = collection.find(targets_query)
            self.logger.info(f"now (target_account_name){target_account_name}")
            if targets.count() == 0:
                continue
            for position in targets:
                        self.logger.info(f"goes in (position){position}")
                        mid = position['mid']
                        query = {'mid' : mid, 'accountName' : target_account_name}
                        holdingdays = position['holding_days']
                        yd_pos = position['actual_td_pos_long'] + position['actual_yd_pos_long']
                        td_pos = 0
                        change = {
                                        'holding_days': holdingdays,
                                        'yd_pos_long': yd_pos,
                                        'td_pos_long': td_pos,
                                        'actual_yd_pos_long': yd_pos,
                                        'actual_td_pos_long': td_pos
                                    }
                        new_data = {"$set": change}
                        res = collection.update_one(
                                                query, new_data, True)
                        self.logger.info(f"[date_change_open] (res){res} (change){change}")

    def update_position_date_close(self):
        for acc in self.accounts_run:
            collection = self.order_info_db[acc]['EquityPosition']
            target_account_name = self.target_account_names[acc]
            targets_query = {'accountName' : target_account_name}
            targets = collection.find(targets_query)
            if targets.count() == 0:
                continue
            for position in targets:
                    mid = position['mid']
                    query = {'mid' : mid, 'accountName': target_account_name}
                    holding_days = position['holding_days'] + 1
                    yd_pos = position['actual_td_pos_long'] + position['actual_yd_pos_long']
                    td_pos = 0
                    change = {
                                        'holding_days': holding_days,
                                        'yd_pos_long': yd_pos,
                                        'td_pos_long': td_pos,
                                        'actual_yd_pos_long': yd_pos,
                                        'actual_td_pos_long': td_pos
                                    }
                    new_data = {"$set": change}
                    res = collection.update_one(
                                                query, new_data, True)
                    self.logger.info(f"[date_change_close] (res){res} (mid){mid} (change){change}")

    def update_atx_order(self, acc, query, update_msg):
        order_info_collection = self.order_info_db[acc]['atx_order']
        order_update_msg = {"$set" : update_msg}
        res = order_info_collection.update_one(query, order_update_msg)
        self.logger.info(f"[update_atx_order] update_atx_order")
        return res







ENCODING = 'gb2312'
GATEWAY_NAME = 'zghs_atx'


class FILES:
    TASKS = 'OrderAlgo_'  #下单
    CANCEL = 'CancelOrderAlgo_' 
    ORDERS = 'ReportOrderAlgo_'
    TRADES = 'SubOrderAlgo_' 
    POSITIONS  = 'ReportPosition_'
    ACCOUNTS = 'ReportBalance_'

#Exchange in tlclient
class Exchange:
    EXCHANGE_UNSPECIFIED = 0
    EXCHANGE_TRADITIONAL = 100 # 表示传统金融
    EXCHANGE_SSE = 101       #  上海证券交易所
    EXCHANGE_SZE = 102


class OrderType:
    ORDER_TYPE_UNSPECIFIED = 0

    ORDER_TYPE_PLAIN_ORDER_PREFIX = 10
    ORDER_TYPE_ALGO_ORDER_PREFIX = 20
    ORDER_TYPE_BASKET_ORDER_PREFIX = 30

    ORDER_TYPE_PLAIN = 100  # 普通单
    ORDER_TYPE_LIMIT = 101  # 限价单
    ORDER_TYPE_MARKET = 102 # 市价单
    ORDER_TYPE_FAK = 103    # 立即成交且剩余撤销（深）
    ORDER_TYPE_FOK = 104    # 立即成交或全部撤销（深）
    ORDER_TYPE_ANY = 105
    ORDER_TYPE_FAK_BEST_5 = 106 # 最优五档立即成交剩余撤销（沪、深）
    ORDER_TYPE_FORWARD_BEST = 107 # 己方最优价（深）
    ORDER_TYPE_REVERSE_BEST = 108 # 对手最优价（深）
    ORDER_TYPE_BEST_5_OR_LIMIT = 109 # 最优五档立即成交剩余转限价（沪）

    ORDER_TYPE_ALGO = 200   # 算法单
    ORDER_TYPE_TWAP = 201   # TWAP算法单
    ORDER_TYPE_VWAP = 202
    ORDER_TYPE_ZC_SMART = 210
    ORDER_TYPE_KF_TWAP_PLUS = 211
    ORDER_TYPE_KF_VWAP_PLUS = 212
    ORDER_TYPE_JN_G_PXINLINE = 213
    ORDER_TYPE_VOLUME_INLINE = 214
    ORDER_TYPE_FT_AIWAP = 215

    ORDER_TYPE_BASKET = 300
    ORDER_TYPE_SINGLE_DAY_BASKET = 301

class EXCHANGE_TYPE:
    SSE = 'SH'
    SZE = 'SZ'


class ORDER_SIDE:
    BUY  = 1  # 股票买入, 或沪港通、深港通股票买入
    SELL = 2  # 股票卖出, 或沪港通、深港通股票卖出
    SIDE_SECURITIES_BUY = 3
    SIDE_FINANCING_SELL = 4
    SIDE_FINANCING_BUY = 5
    SIDE_SECURITIES_SELL = 6

class Side:
    SIDE_UNSPECIFIED = 0
    SIDE_BUY = 1               # 买(担保品买入)
    SIDE_SELL = 2              # 卖(担保品卖出)
    SIDE_FINANCING_BUY = 3     # 融资买入
    SIDE_FINANCING_SELL = 4    # 融券卖出
    SIDE_SECURITIES_BUY = 5    # 买券还券
    SIDE_SECURITIES_SELL = 6   # 卖券还款
    SIDE_SURSTK_TRANS = 7      # 余券划转
    SIDE_STOCK_REPAY_STOCK = 8 # 现券还券

class OrderStatus:
    ORDER_STATUS_UNSPECIFIED = 0
    ORDER_STATUS_UNKNOWN = 1               # 未知
    ORDER_STATUS_INIT = 10                 # 未报
    ORDER_STATUS_PROPOSED = 100            # 已报
    ORDER_STATUS_PENDING_PROPOSE = 101     # 待报
    ORDER_STATUS_PROPOSING = 102           # 正报
    ORDER_STATUS_RESPONDED = 200           # 已响应
    ORDER_STATUS_QUEUEING = 300            # 排队中
    ORDER_STATUS_NO_TRADE_QUEUEING = 301   # 未成交
    ORDER_STATUS_PART_TRADE_QUEUEING = 302 # 部分成交
    ORDER_STATUS_TO_CANCEL = 370           # 待撤销
    ORDER_STATUS_PENDING_MAX = 399

    ORDER_STATUS_REJECTED = 400            # 已拒绝
    ORDER_STATUS_REJECT_BY_ROUTER = 401    # 路由拒单
    ORDER_STATUS_REJECT_BY_GATEWAY = 402   # 网关拒单
    ORDER_STATUS_REJECT_BY_EXCHANGE = 403  # 交易所拒单
    ORDER_STATUS_REJECT_BY_RISK_MGR = 404  # 风控拒单
    ORDER_STATUS_CANCEL_FAILED = 405       # 撤单失败

    ORDER_STATUS_CANCELED = 600            # 已撤销
    ORDER_STATUS_NO_TRADE_CANCELED = 601   # 全部撤单
    ORDER_STATUS_PART_TRADE_CANCELED = 602 # 部成已撤
    ORDER_STATUS_ALL_TRADED = 700          # 全部成交

class SecurityType:
    SECURITY_TYPE_UNSPECIFIED = 0
    SECURITY_TYPE_TRADITIONAL = 100  # 表示传统金融
    SECURITY_TYPE_STOCK = 101        # 股票
    SECURITY_TYPE_FUTURE = 102       # 期货
    SECURITY_TYPE_BOND = 103         # 债券
    SECURITY_TYPE_STOCK_OPTION = 104 # 股票期权
    SECURITY_TYPE_FUND = 105         # 基金
    SECURITY_TYPE_TECH_STOCK = 106   # 科创板
    SECURITY_TYPE_INDEX = 107        # 指数
    SECURITY_TYPE_REPO = 108         # 逆回购
    SECURITY_TYPE_IPO = 109          # 新股申购

    SECURITY_TYPE_CRYPTO = 200       # 数字货币
    SECURITY_TYPE_C_SPOT = 201       # 现货
    SECURITY_TYPE_C_MARGIN = 202     # 现货杠杆
    SECURITY_TYPE_C_FUTURE = 203     # 交割合约
    SECURITY_TYPE_C_SWAP = 204       # 永续合约
    SECURITY_TYPE_C_OPTION = 205     # 期权


class AlgoType:
    ALGO_TYPE = 0
    FT_VWAP_PLUS  = 204 #ft


from datetime import datetime, timezone, timedelta
import os
import queue
import threading
import time
import json
import signal
import sys
import traceback
import argparse
import csv
import pathlib
import shutil
import pandas as pd

from pymongo import MongoClient, ASCENDING, DESCENDING

from ..mongo_client_gateway import MongoClientTradeGateway

from .constants import (GATEWAY_NAME, FILES)
from .utils import (decode_gf_status, side_to_target_type, decode_exchange_id, encode_gf_side,
                   decode_ordtype, encode_gf_market, get_today_date, get_time_from_str, decode_gf_market)



try:
    import thread
except ImportError:
    import _thread as thread

LL9 = 1000000000


class GfServer(MongoClientTradeGateway):
    def __init__(self, config_filename, endtime, gateway_name):
        MongoClientTradeGateway.__init__(self, config_filename, endtime, gateway_name)

        self.load_tg_setting(config_filename)

        self.insert_orders = {}
        self.oid_to_ref = {}
        self.sids = []
        self.oid_to_traded = {}
        self.oid_to_traded_money = {}

        self.sid_to_req = {}
        self.oid_to_req = {}
        self.oid_to_mid = {}
        self.oid_to_acc = {}
        self.db_id_to_oid = {}
        self.ref_to_oid = {}

        self.order_count = 0  # dbf文件当前index
        self.trade_count = 0
        self.oid_to_local_ids = {}
        self.oid_to_instr_ids = {}

    def monitor(self):
        self.thread_pool.submit(self.monitor_algo_order_insert)
        self.thread_pool.submit(self.monitor_cancel_order_insert)
        self.thread_pool.submit(self.monitor_order_update)
        self.thread_pool.submit(self.monitor_trade_update)
        #self.thread_pool.submit(self.update_asset)
        self.thread_pool.submit(self.update_equity_position)
        self.thread_pool.submit(self.date_change)

    def load_tg_setting(self, config_filename):
        try:
            #config_filename = os.path.join(config_path, 'atx_server_config.json')
            f = open(config_filename, encoding='gbk')
            setting = json.load(f)
            #self.insert_order_path = setting['insert_order_path'].replace('/', '\\')
            self.recv_msg_dir = setting['recv_msg_path'].replace('/', '\\')
            self.insert_order_msg_dir = setting['insert_order_msg_path'].replace(
                '/', '\\')
            self.pos_interval = setting['pos_interval']
            self.acc_interval = setting['acc_interval']

        except Exception as e:
            err = traceback.format_exc()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(
                exc_type, exc_value, exc_traceback, limit=None, file=sys.stdout)
            print(f"load config failed! (exception){err}")
            exit(0)

    @MongoClientTradeGateway.error_handler
    def req_buy_order_insert(self, obj):
        current_time = datetime.now()
        formatted_time = current_time.strftime("%Y%m%d%H%M%S%f")[:-3]
        temp_filename = self.insert_order_msg_dir + \
            '_temp\\' + 'buy' + formatted_time + '.csv'

        header = ['local_group_no', 'local_group_name', 'local_report_no', 'projectid', 'market', 'stkcode',
                  'hedgeflag', 'bsflag', 'price', 'qty', 'price_mode', 'price_type', 'diy1', 'diy2', 'algo']

        with open(temp_filename, 'w', newline='') as file:
            writer = csv.writer(file)
            writer.writerow(header)

            target_account_name = obj['accountName']
            acc = self.target_account_names_to_acc[target_account_name]
            account_id = self.account_id[acc]

            oid = str(self.gen_order_id())
            mid = str(obj['mid'])

            local_report_no = oid
            project_id = int(account_id)
            stkcode = obj['ticker']
            market = encode_gf_market(stkcode)
            hedgeflag = '0'
            bsflag = '0B'
            finacial_buy_s = obj["financingBuy"]
            if finacial_buy_s == True:
                bsflag = "4B"
            qty = int(obj["volume"])
            algo_type = decode_ordtype(obj['executionPlan']["order_type"])
            start_time = str(obj['executionPlan']
                             ["start_time"]).replace(":", "")
            end_time = str(obj['executionPlan']["end_time"]).replace(":", "")
            algo_price = 0
            limit_action = 1
            high_limit_buy = 1
            low_limit_sell = 1
            min_amt = 0
            if not obj.get('LimAction') is None :
                if not limit_action:
                    limit_action = 0
                    high_limit_buy = 0
                    low_limit_sell = 0
            after_action = 0
            algo = ''
            if algo_type in ['1001', '1002', '3001', '3002']:
                algo = f"algo_type={algo_type}|start_time={start_time}|end_time={end_time}|algo_price={algo_price}|limit_action={limit_action}|after_action={after_action}"
            elif algo_type in ['4001', '4003']:
                algo = f"algo_type={algo_type}|start_time={start_time}|end_time={end_time}|high_limit_buy={high_limit_buy}|low_limit_sell={low_limit_sell}|after_action={after_action}|min_amt={min_amt}"
            row = ['', '', local_report_no, project_id, market, stkcode,
                   hedgeflag, bsflag, '', qty, '', '', '', '', algo]

            writer.writerow(row)

        # Save the workbook
        current_time = datetime.now()
        formatted_time = current_time.strftime("%Y%m%d%H%M%S%f")[:-3]
        final_filename = f"{self.insert_order_msg_dir}\\instr_001.{formatted_time}.csv"
        shutil.move(temp_filename, final_filename)

        self.logger.info(
            f"[req_buy_order] insert_success (oid){oid} (mid){mid} (param){row}")
        db_id = obj['_id']
        order_dict = {
            'oid': oid,
            'mid': mid,
            "db_id": db_id,
            "ticker": obj['ticker'],
            "volume": qty,
            "accountName": target_account_name,
            "order_type": str(obj['executionPlan']["order_type"]),
            "start_time": str(obj['executionPlan']["start_time"]),
            "end_time": str(obj['executionPlan']["end_time"]),
            "order_time" : datetime.utcnow(),
            'side' : 1
        }

        target_collection = self.order_info_db[acc]['target']
        delete_query = {
            '_id': db_id
        }
        delete_res = target_collection.delete_one(delete_query)
        self.logger.info(f"[on_req_buy_order] delete (target){obj}")

        atx_order_collection = self.order_info_db[acc]['atx_order']
        local_id_list = []
        instr_id_list = []
        db_msg = {
            "oid": oid,
            "mid": mid,
            "traded_vol": 0,
            "traded_amt": 0,
            "local_ids": local_id_list,
            "instr_ids": instr_id_list,
            "order_msg": order_dict
        }
        query = {"oid": oid}
        res = atx_order_collection.replace_one(query, db_msg, True)
        self.logger.info(
            f"[on_req_order_insert] insert_gf_order_info (res){res} (msg){db_msg}")

        self.sids.append(mid)

        self.sid_to_req[mid] = order_dict
        self.oid_to_req[oid] = order_dict
        self.oid_to_mid[oid] = mid
        self.db_id_to_oid[db_id] = oid
        self.oid_to_acc[oid] = acc
        self.oid_to_traded[oid] = 0
        self.oid_to_traded_money[oid] = 0
        self.oid_to_local_ids[oid] = local_id_list
        self.oid_to_instr_ids[oid] = instr_id_list

    @MongoClientTradeGateway.error_handler
    def req_sell_order_insert(self, obj):
        current_time = datetime.now()
        formatted_time = current_time.strftime("%Y%m%d%H%M%S%f")[:-3]
        temp_filename = self.insert_order_msg_dir + \
            '_temp\\' + 'sell' + formatted_time + '.csv'

        header = ['local_group_no', 'local_group_name', 'local_report_no', 'projectid', 'market', 'stkcode',
                  'hedgeflag', 'bsflag', 'price', 'qty', 'price_mode', 'price_type', 'diy1', 'diy2', 'algo']

        with open(temp_filename, 'w', newline='') as file:
            writer = csv.writer(file)
            writer.writerow(header)

            target_account_name = obj['accountName']
            acc = self.target_account_names_to_acc[target_account_name]
            account_id = self.account_id[acc]

            oid = str(self.gen_order_id())
            mid = str(obj['mid'])

            local_report_no = oid
            project_id = int(account_id)
            stkcode = obj['ticker']
            market = encode_gf_market(stkcode)
            hedgeflag = '0'
            bsflag = '0S'
            qty = int(obj["volume"])
            algo_type = decode_ordtype(obj['executionPlan']["order_type"])
            start_time = str(obj['executionPlan']
                             ["start_time"]).replace(":", "")
            end_time = str(obj['executionPlan']["end_time"]).replace(":", "")
            algo_price = 0
            limit_action = 1
            if not obj.get('LimAction') is None :
                if not limit_action:
                    limit_action = 0


            after_action = 0
            algo = f"algo_type={algo_type}|start_time={start_time}|end_time={end_time}|algo_price={algo_price}|limit_action={limit_action}|after_action={after_action}"
            row = ['', '', local_report_no, project_id, market, stkcode,
                   hedgeflag, bsflag, '', qty, '', '', '', '', algo]

            writer.writerow(row)

        # Save the workbook
        current_time = datetime.now()
        formatted_time = current_time.strftime("%Y%m%d%H%M%S%f")[:-3]
        final_filename = f"{self.insert_order_msg_dir}\\instr_001.{formatted_time}.csv"
        shutil.move(temp_filename, final_filename)

        self.logger.info(
            f"[req_sell_order] insert_success (oid){oid} (mid){mid} (param){row}")
        db_id = obj['_id']
        order_dict = {
            'oid': oid,
            'mid': mid,
            "db_id": db_id,
            "ticker": obj['ticker'],
            "volume": qty,
            "accountName": target_account_name,
            "order_type": str(obj['executionPlan']["order_type"]),
            "start_time": str(obj['executionPlan']["start_time"]),
            "end_time": str(obj['executionPlan']["end_time"]),
            "order_time" : datetime.utcnow(),
            "side" : 2
        }

        target_collection = self.order_info_db[acc]['sell_target']
        delete_query = {
                    "_id" : db_id
                }
        delete_res = target_collection.delete_one(delete_query)
        self.logger.info(f"[on_req_sell_order] delete (target){delete_res}")

        atx_order_collection = self.order_info_db[acc]['atx_order']
        local_id_list = []
        instr_id_list = []
        db_msg = {
            "oid": oid,
            "mid": mid,
            "traded_vol": 0,
            "traded_amt": 0,
            "local_ids": local_id_list,
            "instr_ids": instr_id_list,
            "order_msg": order_dict
        }
        query = {"oid": oid}
        res = atx_order_collection.replace_one(query, db_msg, True)
        self.logger.info(
            f"[on_req_order_insert] insert_gf_order_info (res){res} (msg){db_msg}")

        self.sids.append(mid)

        self.sid_to_req[mid] = order_dict
        self.oid_to_req[oid] = order_dict
        self.oid_to_mid[oid] = mid
        self.db_id_to_oid[db_id] = oid
        self.oid_to_acc[oid] = acc
        self.oid_to_traded[oid] = 0
        self.oid_to_traded_money[oid] = 0
        self.oid_to_local_ids[oid] = local_id_list
        self.oid_to_instr_ids[oid] = instr_id_list


    @MongoClientTradeGateway.error_handler
    def monitor_order_update(self):
        while not self.is_stopped:

            order_filename = self.recv_msg_dir + '\\' + \
                FILES.ORDERS + '.' + get_today_date() + '.csv'
            
            data_list = []

            df = pd.read_csv(order_filename, dtype={'local_report_no':str})

            header = ['filename', 'local_group_no', 'local_group_name', 'local_report_no', 'instr_no', 'instrstk_no',  'projectid', 'market', 'stkcode', 'hedgeflag', 'bsflag', 'orderqty',
                      'orderprice', 'orderstatus', 'matchqty', 'matchamt', 'cancelqty', 'fees', 'accrued_interest', 'clearamt', 'ordertime', 'direct_operid', 'order_operid', 'remark', 'ordersno', 'orderextno']

            data_list = []
            current_len = df.shape[0]
            
            if(current_len > self.order_count):
                for _, row in df.iloc[self.order_count:].iterrows():
                    row_dict = {}
                    for col in header:
                        row_dict[col] = row[col]
                    data_list.append(row_dict)

                for record in data_list:
                    if record['orderstatus'] in ['5', '8']:
                        self.update_order(record)
                    elif record['orderstatus'] in ['9']:
                        self.logger.info(
                            f"[monitor_order_update] order_rejected msg{record}")
                    
                self.order_count = current_len

            time.sleep(self.order_scan_interval)

    @MongoClientTradeGateway.error_handler
    def update_asset(self):
        while not self.is_stopped:
            asset_filename = self.recv_msg_dir + '\\' + \
                FILES.ACCOUNTS + '.' + get_today_date() + '.csv'
            df = pd.read_csv(asset_filename, dtype={'projectid': int})
            for _, row in df.iterrows():
                account_id = row["projectid"]
                instravl = row["instravl"]
                hkinstravl = row["hkinstravl"] #港股可用资金
                netasset = row["netasset"]
                stkasset = row["stkasset"]
                ftmargin = row["ftmargin"] #持仓占用保证金
                creditmargin = row["ftmargin"] #保证金可用
                assurebalance = row["assurebalance"] #担保品可用
                all_asset = row["allasset"]
                acc = ""
                try:
                    acc = self.account_id_to_acc[int(account_id)]
                except:
                    self.logger.warning(f"account {int(account_id)} not using")
                    continue
                tg_name = self.tgnames[acc]
                accountname = self.log_account_names[acc]
                db_msg = {
                    "_id": tg_name,  # 对应之前的 TG 名称
                    "accountName": accountname,  # 产品名称
                    "avail_amt": instravl,
                    "balance": netasset,
                    "holding": stkasset,
                    'guaranty': creditmargin,
                    "all_asset": all_asset,
                    "updated_at": datetime.utcnow()
                }

                update_msg = {"$set" : db_msg}
                asset_collection = self.order_info_db[acc]['TestEquityAccount']
                query = {'_id': tg_name, 'accountName': accountname}
                res = asset_collection.update_one(query, update_msg)
                self.logger.info(f"[rtn_asset] (res){res} (asset_msg){db_msg}")
            
            time.sleep(self.acc_interval)
     
    @MongoClientTradeGateway.error_handler           
    def update_equity_position(self):
        while not self.is_stopped:
            pos_filename = self.recv_msg_dir + '\\' + \
                FILES.POSITIONS + '.' + get_today_date() + '.csv'
            dtypes ={
                'projectid': int,
                'stkcode' : str,
                'market': str,
                'direction' : str
            }
            df = pd.read_csv(pos_filename, dtype=dtypes)
            for acc in self.accounts_run:
                tg_position_collection = self.order_info_db[acc]['tg_equity_position']
                remove = tg_position_collection.delete_many({'account_name': self.log_account_names[acc], 'tg_name': self.target_account_names[acc]})
                self.logger.info(f"[update_equity_position] delete_old_position_info (remove){remove} ")
            for _, row in df.iterrows():
                account_id = row["projectid"]
                market = row["market"]
                exchange = decode_gf_market(market)
                stkcode = int(row["stkcode"])
                ticker = f'{stkcode:06d}'
                direction = row["direction"] # 'B' 多 ‘S’ 空
                out_direction = "long"
                if direction == 'S':
                    out_direction = 'short'
                schemeid = row["schemeid"]
                hedgeflag = row["hedgeflag"]
                stkholdqty = row["stkholdqty"] #
                stkavl = row["stkavl"] #
                costprice = row["costprice"] #
                #ftmargin = row["ftmargin"]

                acc = ""
                try:
                    acc = self.account_id_to_acc[int(account_id)]
                except:
                    self.logger.warning(f"account {int(account_id)} not using")
                    continue
                tg_name = self.target_account_names[acc]
                accountname = self.log_account_names[acc]
                pos_collection = self.order_info_db[acc]['tg_equity_position']
                query = {'tg_name':tg_name, 'accountName': accountname, "ticker":ticker}
                pos_msg = {
                    "account_name": accountname,
                    "tg_name": tg_name,
                    "ticker": ticker,
                    "exchange": exchange,
                    "direction": out_direction, # long/short，没有short部分就只存long部分
                    "avail_pos": stkavl, # 昨仓
                    "total_pos": stkholdqty, # 今仓，TODO 需要统计下不同券商盘中，对于卖出的position 是直接从 yd_pos 上减，还是在 td_pos 增加一个负的值。
                    "cost": costprice, #没有
                    "type": "stock",
                    "updated_at": datetime.utcnow()
                                }
                res = pos_collection.replace_one(pos_msg, pos_msg, True)
                self.logger.info(f"[update_equity_position] (res){res.modified_count} (order_msg){pos_msg}")
            time.sleep(self.pos_interval)

    @MongoClientTradeGateway.error_handler
    def update_order(self, record):
        if record['matchqty'] == 0:
            self.logger.warning(
                f"[update_order] nothing_traded (record){record}")
            return

        oid = str(record['local_report_no'])
        if  len(oid) == 0:
            self.logger.warning(f"[update_order] not use file to trade record:{record}")
            return

        acc = ""
        mid = ""
        order_dict = {}
        mudan_id = str(record['instrstk_no'])
        if not oid in self.oid_to_acc:  # shut_down or other problem
            trade_acc = int(record['projectid'])

            if trade_acc not in self.account_id_to_acc:
                self.logger.error(
                    f"[update_order] can't_parse_trade_acc {trade_acc}")
                return
            acc = self.account_id_to_acc[trade_acc]
            order_info_collection = self.order_info_db[acc]['atx_order']
            query = {'oid': oid}
            order_info_target = order_info_collection.find_one(query)
            if not order_info_target is None:
                order_dict = order_info_target['order_msg']
                mid = order_info_target['mid']
                try:
                    self.oid_to_instr_ids[oid] = order_info_target['instr_ids']
                except:
                    self.oid_to_instr_ids[oid] = []
                self.oid_to_traded_money[oid] = 0
                self.oid_to_traded[oid] = 0
                self.oid_to_mid[oid] = mid
                self.oid_to_req[oid] = order_dict
                self.oid_to_ref[oid] = mudan_id
            else:
                self.logger.error(f"[update_order] can't_find_oid (oid){oid}")
                return
        else:
            acc = self.oid_to_acc[oid]
            order_dict = self.oid_to_req[oid]
            mid = self.oid_to_mid[oid]
            

        instr_id = str(record['instrstk_no'])
        if instr_id in self.oid_to_instr_ids[oid]:
            return
        else:
            self.oid_to_instr_ids[oid].append(instr_id)
        traded_vol = record['matchqty']

        int_stock_code = int(record['stkcode'])
        ticker = f'{int_stock_code:06d}'
        exchange = decode_exchange_id(ticker)
        volume = record['orderqty']
        
        price = 0
        if(traded_vol != 0):
            price = float(record['matchamt'] / traded_vol)

        tg_name = order_dict['accountName'].split('@')[0]
        accountName = order_dict['accountName'].split('@')[1]
        order_type = order_dict['order_type']
        start_time = order_dict['start_time']
        end_time = order_dict['end_time']
        order_time = order_dict['order_time']

        target_type = side_to_target_type(order_dict['side'])
        status = decode_gf_status(record['orderstatus'])
        utc_start_time = get_time_from_str(str(start_time))

        replace_collection = self.tradelog_db[acc]['order']
        query = {'mid': mid, '_id': int(oid)}
        replace_target = replace_collection.find_one(query)
        dbtime = order_time if order_time > utc_start_time else utc_start_time

        if replace_target is None:
            self.oid_to_ref[oid] = mudan_id
            order_msg = {
                "_id": int(oid),
                "tg_name": tg_name,  # 对应 account_info._id
                "exchange": exchange, # 对应 tlclient.trader.constant 包 ExchangeID
                "target_type": target_type,    # 'buy' | 'sell' | 'limit_sell'
                "volume": volume,  # 订单的volume
                "price": price,  # 实际成交均价
                "order_type": order_type, 
                "ticker": ticker,
                "mid": mid,  # target中对应的 母单sid
                "accountName": accountName,  # 对应 account_info.account_name
                "algo_args": {  # 具体的算法参数
                    "order_type": order_type,  # 此柜台不使用下单时要求的算法
                    "start_time": start_time,
                    "end_time": end_time
                },
                "status": status,  # 'active' | 'filled' | 'canceled'
                "filled_vol": traded_vol,  # 实际成交的 volume
                "dbTime": dbtime,
            }

            res = replace_collection.replace_one(query, order_msg, True)
            self.logger.info(f"[rtn_order] (res){res} (order_msg){order_msg}")
        else:
            price = (replace_target['filled_vol'] * replace_target['price'] + price * traded_vol) / (traded_vol + replace_target['filled_vol'])
            volume += replace_target['volume']
            traded_vol += replace_target['filled_vol']
            
            order_msg = {
                "_id": int(oid),
                "tg_name": tg_name,  # 对应 account_info._id
                "exchange": exchange, # 对应 tlclient.trader.constant 包 ExchangeID
                "target_type": target_type,    # 'buy' | 'sell' | 'limit_sell'
                "volume": volume,  # 订单的volume
                "price": price,  # 实际成交均价
                "order_type": order_type, 
                "ticker": ticker,
                "mid": mid,  # target中对应的 母单sid
                "accountName": accountName,  # 对应 account_info.account_name
                "algo_args": {  # 具体的算法参数
                    "order_type": order_type,  # 此柜台不使用下单时要求的算法
                    "start_time": start_time,
                    "end_time": end_time
                },
                "status": status,  # 'active' | 'filled' | 'canceled'
                "filled_vol": traded_vol,  # 实际成交的 volume
                "dbTime": dbtime,
            }

            res = replace_collection.replace_one(query, order_msg, True)
            self.logger.info(f"[rtn_order] (res){res} (order_msg){order_msg}")
            
        atx_order_collection = self.order_info_db[acc]['atx_order']
        atx_order_msg = {
            "instr_ids": self.oid_to_instr_ids[oid],
        }
        update_atx_order_msg = {"$set": atx_order_msg}
        query = {"oid": oid}
        res = atx_order_collection.update_one(query, update_atx_order_msg)
        self.logger.info(
            f"[update_order] update_atx_order_info (res){res} (msg){update_atx_order_msg}")


    @MongoClientTradeGateway.error_handler
    def monitor_algo_order_insert(self):
        while not self.is_stopped:
            print("[monitor_algo_order_insert]")
            for acc in self.accounts_run:
                target_account_name = self.target_account_names[acc]
                buy_query = {"accountName": target_account_name}
                buy_targets = self.order_info_db[acc]["target"].find(buy_query)
                if buy_targets.count() == 0:
                    self.logger.warning(
                        f"[monitor_buy_order] no_buy_target (acc){acc}")
                    continue
                for target in buy_targets:
                    if target['_id'] not in self.order_db_ids:
                        self.order_db_ids.append(target['_id'])
                        self.req_buy_order_insert(target)
                    else:
                        self.logger.warning(
                            f"[monitor_algo_order_insert] _id_existed (_id){target['_id']}")
            for acc in self.accounts_run:
                target_account_name = self.target_account_names[acc]
                sell_query = {"accountName": target_account_name}
                sell_targets = self.order_info_db[acc]["sell_target"].find(
                    sell_query)
                if sell_targets.count() == 0:
                    self.logger.warning(
                        f"[monitor_sell_order] no_sell_target (acc){acc}")
                    continue
                for sell_target in sell_targets:
                    if sell_target['_id'] not in self.sell_order_db_ids:
                        self.sell_order_db_ids.append(sell_target['_id'])
                        self.req_sell_order_insert(sell_target)
                    else:
                        self.logger.warning(
                            f"[monitor_algo_order_insert] _id_existed (_id){sell_target['_id']}")
            time.sleep(self.scan_interval)

    @MongoClientTradeGateway.error_handler
    def monitor_cancel_order_insert(self):
        while not self.is_stopped:
            for acc in self.accounts_run:
                print("[monitor_cancel_order_insert]")
                target_account_name = self.target_account_names[acc]
                query = {"accountName": target_account_name}
                with self.cancel_orderlock:
                    cancel_targets = self.order_info_db[acc]["cancel_target"].find(
                        query)
                    if cancel_targets.count() == 0:
                        self.logger.warning(
                            "[monitor_cancel_order] no_cancel_info")
                        continue
                    for cancel_target in cancel_targets:
                        if cancel_target['_id'] not in self.cancel_order_ids:
                            self.cancel_order_ids.append(cancel_target['_id'])
                            self.req_cancel_order(cancel_target, acc)
            time.sleep(self.scan_interval)

    @MongoClientTradeGateway.error_handler
    def req_cancel_order(self, cancel_target, acc):
        current_time = datetime.now()
        formatted_time = current_time.strftime("%Y%m%d%H%M%S%f")[:-3]
        temp_filename = f"{self.insert_order_msg_dir}_temp\\cancel_001.{formatted_time}.csv"

        header = ['local_cancel_no', 'group_flag',
                  'local_no', 'projectid', 'diy1', 'diy2']

        with open(temp_filename, 'w', newline='') as file:
            writer = csv.writer(file)
            writer.writerow(header)
            oid = str(cancel_target['oid'])
            account_id = self.account_id[acc]
            mudan_id = ""
            if oid not in self.oid_to_ref:
                self.logger.error(
                    "[req_cancel_order] can't_find (oid){oid}")
                return
            mudan_id = self.oid_to_ref[oid]
            group_flag = 0
            cancel_id = str(self.gen_order_id())
            row = [cancel_id, group_flag, oid, account_id, '', '']
            writer.writerow(row)

        final_filename = f"{self.insert_order_msg_dir}\\cancel_001.{formatted_time}.csv"
        shutil.move(temp_filename, final_filename)

        origin_req = self.oid_to_req[oid]
        target_account_name = origin_req['accountName']
        acc = self.target_account_names_to_acc[target_account_name]
        target_collection = self.order_info_db[acc]['cancel_target']
        delete_query = {
            'accountName': target_account_name,
            'oid': oid
        }
        delete_res = target_collection.delete_one(delete_query)
        self.logger.info(
            f"[on_req_cancel_order] delete (target){cancel_target}")

    @MongoClientTradeGateway.error_handler
    def monitor_trade_update(self):
        while not self.is_stopped:
            print("[monitor_trade_update]")
            trade_filename = self.recv_msg_dir + '\\' +  FILES.TRADES + '.' + get_today_date() + '.csv'
            data_list = []

            df = pd.read_csv(trade_filename, dtype={'local_report_no':str})

            header = ['filename', 'local_group_no', 'local_group_name', 'local_report_no', 'instr_no', 'instrstk_no', 'match_no', 'projectid', 'stkcode', 'hedgeflag', 'bsflag', 'matchtime',
                      'matchtype', 'matchprice', 'matchqty', 'matchamt', 'direct_operid', 'order_operid', 'remark', 'matchsno', 'orderextno']

            data_list = []
            current_len = df.shape[0]

            if(current_len > self.trade_count):
                for _, row in df.iloc[self.trade_count:].iterrows():
                    row_dict = {}
                    for col in header:
                        row_dict[col] = row[col]
                    data_list.append(row_dict)

                for record in data_list:
                    if int(record['matchtype']) in [0]:
                        self.on_trade_msg(record)
                    else:
                        self.logger.info(
                            f"[monitor_trade_update] trade_not_finished_or_rejected_or_canceled msg{record}")
                self.trade_count = current_len

            time.sleep(self.trade_scan_interval)

    @MongoClientTradeGateway.error_handler
    def on_trade_msg(self, record):
        if record['matchqty'] == 0:
            self.logger.warning(
                f"[on_trade_msg] nothing_traded (record){record}")
            return

        oid = str(record['local_report_no'])
        if  len(oid) == 0:
            self.logger.warning(f"[on_trade_msg] not use file to  trade record:{record}")
            return

        mudan_id = str(record['instrstk_no'])

        acc = ""
        mid = ""
        order_dict = {}
        if not oid in self.oid_to_acc:  # shut_down or other problem
            trade_acc = record['projectid']
            acc = self.account_id_to_acc[trade_acc]
            order_info_collection = self.order_info_db[acc]['atx_order']
            query = {'oid': oid}
            order_info_target = order_info_collection.find_one(query)

            if not order_info_target is None:
                order_dict = order_info_target['order_msg']
                mid = order_info_target['mid']
                self.oid_to_local_ids[oid] = order_info_target['local_ids']
                self.oid_to_traded[oid] = 0
                self.oid_to_traded_money[oid] = 0
                self.oid_to_mid[oid] = mid
                self.oid_to_req[oid] = order_dict
                self.oid_to_ref[oid] = mudan_id
            else:
                self.logger.error(f"[on_trade_msg] can't_find_oid (oid){oid}")
                return
        else:
            acc = self.oid_to_acc[oid]
            order_dict = self.oid_to_req[oid]
            mid = self.oid_to_mid[oid]
        _id = str(int(record['match_no']))
        if _id in self.oid_to_local_ids[oid]:
            #self.logger.warning(f"[on_trade_msg] duplicate_trade_msg (record){record}")
            return
        else:
            self.oid_to_local_ids[oid].append(_id)
        int_stock_code = int(record['stkcode'])
        ticker = f'{int_stock_code:06d}'
        exchange = decode_exchange_id(ticker)
        traded_vol = int(record['matchqty'])
        traded_price = float(record['matchprice'])
        trade_amt = record['matchamt']

        target_type = encode_gf_side(record['bsflag'])
        if target_type == 1 or target_type == 3:
            self.oid_to_traded_money[oid] += trade_amt
            self.oid_to_traded[oid] += traded_vol
        elif target_type == 2:
            self.oid_to_traded_money[oid] -= trade_amt
            self.oid_to_traded[oid] -= traded_vol
        entrust_vol = record['matchqty']
        #entrust_price = record.Price
        entrust_price = float(record['matchprice'])
        dbTime = get_time_from_str(str(record['matchtime']))

        commission = 0
        order_type = order_dict['order_type']
        target_account_name = order_dict['accountName']
        acc = self.target_account_names_to_acc[target_account_name]
        log_account_name = order_dict['accountName'].split('@')[1]
        tg_name = order_dict['accountName'].split('@')[0]
        side = encode_gf_side(record['bsflag'])
        trade_collection = self.tradelog_db[acc]['tg_trade']
        replace_trade_query = { "trade_ref": _id, "oid" : int(oid), "mid": mid, "accountName": log_account_name}
        trade_target = trade_collection.find_one(replace_trade_query)
        if not trade_target is None:
            dbTime = trade_target['dbTime']

        db_msg = {
            "trade_ref": _id,
            "oid": int(oid),
            "tg_name": tg_name,
            "exchange": exchange,
            "ticker": ticker,
            "traded_vol": traded_vol,
            "traded_price": traded_price,
            "order_type": order_type,
            "side": side,  # 对应 tlclient.trader.constant 包 Side
            "entrust_vol": entrust_vol,
            "entrust_price": entrust_price,  # 没有下单价格，只有成交价
            "dbTime": dbTime,
            "mid": mid,  # 对应订单中的 mid
            "commission": commission,  # 没有
            "trade_time": dbTime,  # 具体交易时间

            "accountName":  log_account_name,  # 对应 account_info.account_name
        }
        db_res = trade_collection.replace_one(
            replace_trade_query, db_msg, True)
        self.logger.info(
            f"[rtn_trade] (db_res){db_res} (db_msg){db_msg} (traded_vol){traded_vol} (traded_price){traded_price}")
        #self.update_position(traded_vol, order_dict, mid,oid, side, trade_amt, exchange)
        atx_order_collection = self.order_info_db[acc]['atx_order']
        atx_order_msg = {
            "local_ids": self.oid_to_local_ids[oid],
        }
        update_atx_order_msg = {"$set": atx_order_msg}
        query = {"oid": oid}
        res = atx_order_collection.update_one(query, update_atx_order_msg)
        self.logger.info(
            f"[on_trade_msg] update_atx_order_info (res){res} (msg){update_atx_order_msg}")

    @MongoClientTradeGateway.error_handler
    def update_position(self, traded_vol, order, mid, oid, side, trade_amt, exchange):
        target_account_name = order['accountName']
        acc = self.target_account_names_to_acc[target_account_name]

        query = {'mid': mid, 'accountName': target_account_name}
        position_collection = self.order_info_db[acc]['EquityPosition']
        position = position_collection.find_one(query)
        if position is None and traded_vol > 0:
            if oid not in self.oid_to_req:
                self.logger.warning(
                    f"[update_position] can't_find_req (mid){mid} (oid){oid}")
                return
            if oid not in self.oid_to_traded:
                self.logger.error(f"[update_position]no trade_vol (mid){mid}")
                return
            td_trade_vol = self.oid_to_traded[oid]

            amt = self.oid_to_traded_money[oid]

            ticker = order['ticker']
            yd_pos = 0
            mkt_value = amt
            cost = 0
            if not td_trade_vol == 0:
                cost = amt / td_trade_vol
            update_time = datetime.now(timezone.utc)

            position_msg = {
                "ticker": ticker,
                "cost": cost,  # 成本价
                "td_pos_long": td_trade_vol,  # 今仓
                "yd_pos_long": yd_pos,  # 昨仓（可卖），挂止盈单后为0
                "td_pos_short": 0,
                "yd_pos_short": 0,
                "actual_td_pos_long": td_trade_vol,  # 实际的今仓
                "actual_yd_pos_long": yd_pos,  # 实际的昨仓
                "enter_date": update_time,
                "holding_days": 0,  # 持仓时间（交易日累计）
                "mkt_value": mkt_value,  # 市值
                "exchange": exchange,  # 对应 tlclient.trader.constant 包 ExchangeID
                "mid": mid,  # 我们策略生成的id，需要根据实际下单情况维护各sid的持仓。
                # 格式: {account_info.tg_name}@{account_info.account_name}
                "accountName": target_account_name,
                "update_date": update_time,
            }
            query2 = {'mid': mid,
                      'accountName': target_account_name}
            res = position_collection.replace_one(query2, position_msg, True)
            self.logger.info(
                f"[update_position] (res){res} (msg){position_msg}")
        else:
            if oid not in self.oid_to_req:
                self.logger.warning(
                    f"[update_position] can't_find_req (mid){mid}")
                return
            if oid not in self.oid_to_traded:
                self.logger.error(f"[update_position]no trade_vol (mid){mid}")
                return
            actual_yd_pos = position['actual_yd_pos_long']
            amt = trade_amt
            yd_pos = position['yd_pos_long']
            if side == 1 or side == 3:
                td_trade_vol = position['actual_td_pos_long'] + traded_vol
                mkt_value = position['mkt_value'] + amt
            elif side == 2:
                td_trade_vol = position['actual_td_pos_long'] - traded_vol
                mkt_value = position['mkt_value'] - amt
                yd_pos = yd_pos - traded_vol

            ticker = order['ticker']

            cost = 0
            total_vol = td_trade_vol + yd_pos
            if not total_vol == 0:
                cost = mkt_value / total_vol
            update_time = datetime.now(timezone.utc)
            accountName = order['accountName']
            enter_time = position['enter_date']
            holding_days = position['holding_days']
            position_msg = {
                "ticker": ticker,
                "cost": cost,  # 成本价
                "td_pos_long": td_trade_vol,  # 今仓
                "yd_pos_long": yd_pos,  # 昨仓（可卖），挂止盈单后为0
                "td_pos_short": 0,
                "yd_pos_short": 0,
                "actual_td_pos_long": td_trade_vol,  # 实际的今仓
                "actual_yd_pos_long": actual_yd_pos,  # 实际的昨仓
                "enter_date": enter_time,
                "holding_days": holding_days,  # 持仓时间（交易日累计）
                "mkt_value": mkt_value,  # 市值
                "exchange": exchange,  # 对应 tlclient.trader.constant 包 ExchangeID
                "mid": mid,  # 我们策略生成的id，需要根据实际下单情况维护各sid的持仓。
                # 格式: {account_info.tg_name}@{account_info.account_name}
                "accountName": accountName,
                "update_date": update_time,
            }
            query = {'mid': mid,
                     'accountName': accountName}
            res = position_collection.replace_one(query, position_msg, True)
            self.logger.info(
                f"[update_position] (res){res} (msg){position_msg}")

    def start(self):
        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGINT, self.signal_handler)
        _msg = f"[login] gf_server_start (time){datetime.now()}"
        self.send_to_user(logger=self.logger, url_list=self.url_list, msg=_msg)
        self.monitor()

    def close(self):
        return super().close()

    def join(self):
        while self.is_stopped == False:
            time.sleep(0.01)
            if self.is_stopped:
                self.logger.info(
                    "[close] main thread is stopped,active_orders message will lose")
                self.close()


if __name__ == "__main__":

    sys.stdout.reconfigure(encoding='utf-8')
    description = "gf_server,get target from mongodb and serve gf"
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument('-e', '--end_time', dest='end_time', default='15:30')
    _config_filename = "C:/Users/Administrator/Desktop/gf_batandconfig/gf_server_config.json"
    parser.add_argument('-p', '--config_filepath',
                        dest='config_filepath', default=_config_filename)

    args = parser.parse_args()
    print(f"(args){args}")

    td = GfServer(args.config_filepath, args.end_time, GATEWAY_NAME)
    td.start()

    td.join()
